"""Interface specification for network event detector backends."""

from abc import abstractmethod, ABCMeta
from contextlib import asynccontextmanager
from typing import AsyncIterator, Any, Dict, List

__all__ = ("NetworkEventDetectorBackend",)

NetworkInterface = Any


class NetworkEventDetectorBackend(metaclass=ABCMeta):
    """Interface specification for network event detector backends."""

    def configure(self, configuration: Dict[str, Any]) -> None:
        """Configures the detector backend and specifies what the backend should
        report.

        The format of the configuration dictionary depends on the backend.
        The default implementation does nothing.

        It is guaranteed that no one else holds a reference to the configuration
        dictionary so it is safe to just store it as-is (without making a
        copy first).
        """
        pass  # pragma: no cover

    @abstractmethod
    async def get_addresses(self, interface: NetworkInterface) -> Dict[int, List[Any]]:
        """Returns all known addresses for a given network interface.

        Returns:
            a dictionary mapping address family identifiers (typically numbers)
            to lists of corresponding network addresses. It is guaranteed to be
            an instance that the caller can freely modify.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def key_of(self, interface: NetworkInterface) -> str:
        """Returns a unique key for a network interface that can be used for
        identity comparisons.

        The string does not have to be human-readable, but it must be unique
        for each network interface.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    async def scan(self) -> List[NetworkInterface]:
        """Scans the system for network interfaces and returns the list of
        interfaces found.
        """
        raise NotImplementedError  # pragma: no cover

    @asynccontextmanager
    async def use(self) -> AsyncIterator[None]:
        """Async context manager that is entered when the backend starts up and
        exits when the backend shuts down.

        This context manager can be used to run background tc_genhreads or manage
        any additional resources that correspond to the lifetime of the backend.

        The default implementation does nothing.
        """
        yield

    @abstractmethod
    async def wait_until_next_scan(self) -> None:
        """Async function that blocks until the next scan is due.

        It is up to the backend to decide whether it scans the network
        interfaces regularly or whether it uses some underlying mechanism of the
        OS to detect events.
        """
        raise NotImplementedError  # pragma: no cover



"""Async event detector for network interfaces."""

from anyio import Event
from contextlib import contextmanager
from enum import Enum
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from .backends.base import NetworkEventDetectorBackend, NetworkInterface
from .backends.autodetect import choose_backend


__all__ = ("NetworkEventDetector",)


class NetworkEventType(Enum):
    """Enum representing the possible network event types."""

    INTERFACE_ADDED = "interface_added"
    INTERFACE_REMOVED = "interface_removed"
    ADDRESS_ADDED = "address_added"
    ADDRESS_REMOVED = "address_removed"


NetworkEvent = NamedTuple(
    "NetworkEvent",
    [
        ("type", NetworkEventType),
        ("interface", NetworkInterface),
        ("address_family", Optional[int]),
        ("address", Any),
        ("key", str),
    ],
)
NetworkEventDetectorEntry = NamedTuple(
    "NetworkEventDetectorEntry",
    [
        ("interface", NetworkInterface),
        ("key", str),
        ("addresses", Dict[int, List[Any]]),
    ],
)


class NetworkEventDetector:
    """Network event detector class.

    This class continuously scans the system for network interfaces and dispatches
    events when network interfaces are added or removed, or when a network
    interface gains a new IP address or loses an existing one.
    """

    _backend: Union[
        NetworkEventDetectorBackend, Callable[[], NetworkEventDetectorBackend]
    ]
    _entries: Dict[str, NetworkEventDetectorEntry]
    _params: Dict[str, Any]
    _resume_event: Optional[Event]
    _suspended: int

    def __init__(
        self,
        params: Optional[Dict] = None,
        *,
        backend: Callable[[], NetworkEventDetectorBackend] = choose_backend
    ):
        """Constructor.

        Parameters:
            params: dictionary of keyword arguments to pass to the
                `configure()` method of the backend in order to specify what
                sort of interfaces we are interested in.
            backend: a callable that returns an instance of NetworkEventScannerBackend_
                when invoked with no arguments; defaults to autodetection
                depending on the current platform.
        """
        self._backend = backend

        self._params = self._preprocess_params(dict(params or {}))
        self._entries = {}

        self._suspended = 0
        self._resume_event = None

    async def added_addresses(self) -> AsyncIterator[NetworkEvent]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            event objects describing the network addresses that were added and
            the corresponding network interfaces
        """
        async for event in self.events():
            if event.type == NetworkEventType.ADDRESS_ADDED:
                yield event

    async def added_interfaces(self) -> AsyncIterator[NetworkInterface]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            network interface objects, one for each interface whose addition was
            detected. Removed network interfaces are not reported.
        """
        async for event in self.events():
            if event.type == NetworkEventType.INTERFACE_ADDED:
                yield event.interface

    async def changed_addresses(self) -> AsyncIterator[NetworkEvent]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            event objects describing the network addresses that were added or
            removed and the corresponding network interfaces
        """
        async for event in self.events():
            if event.type in (
                NetworkEventType.ADDRESS_ADDED,
                NetworkEventType.ADDRESS_REMOVED,
            ):
                yield event

    async def events(self) -> AsyncIterator[NetworkEvent]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            event objects describing the network interfaces or addresses that
            were added or removed
        """
        backend = self._backend() if callable(self._backend) else self._backend
        backend.configure(self._params)
        key_of = backend.key_of

        async with backend.use():
            while True:
                if self._suspended:
                    assert self._resume_event is not None
                    await self._resume_event.wait()

                interfaces = await backend.scan()

                added = {}
                changed = {}

                for index, interface in enumerate(interfaces):
                    addresses = await backend.get_addresses(interface)
                    key = key_of(interface)

                    entry = self._entries.get(key)
                    if entry is not None:
                        changed[key] = self._compare_addresses(entry, addresses)
                        if any(changed[key]):
                            self._entries[key] = entry = entry._replace(
                                addresses=addresses
                            )
                    else:
                        entry = added[key] = NetworkEventDetectorEntry(
                            interface=interface, addresses=addresses, key=key
                        )

                to_remove = [key for key in self._entries.keys() if key not in changed]
                removed = [self._entries.pop(key) for key in to_remove]
                self._entries.update(added)

                for entry in removed:
                    interface = entry.interface
                    key = entry.key

                    for family, addresses in entry.addresses.items():
                        for address in addresses:
                            event = NetworkEvent(
                                type=NetworkEventType.ADDRESS_REMOVED,
                                interface=interface,
                                key=key,
                                address=address,
                                address_family=family,
                            )
                            yield event

                    event = NetworkEvent(
                        type=NetworkEventType.INTERFACE_REMOVED,
                        interface=interface,
                        key=key,
                        address=None,
                        address_family=None,
                    )
                    yield event

                for key, (added_entries, removed_entries) in changed.items():
                    for entry in removed_entries:
                        interface, key, family, address = entry
                        event = NetworkEvent(
                            type=NetworkEventType.ADDRESS_REMOVED,
                            interface=interface,
                            key=key,
                            address=address,
                            address_family=family,
                        )
                        yield event

                    for entry in added_entries:
                        interface, key, family, address = entry
                        event = NetworkEvent(
                            type=NetworkEventType.ADDRESS_ADDED,
                            interface=interface,
                            key=key,
                            address=address,
                            address_family=family,
                        )
                        yield event

                for entry in added.values():
                    interface = entry.interface
                    key = entry.key

                    event = NetworkEvent(
                        type=NetworkEventType.INTERFACE_ADDED,
                        interface=interface,
                        key=key,
                        address=None,
                        address_family=None,
                    )
                    yield event

                    for family, addresses in entry.addresses.items():
                        for address in addresses:
                            event = NetworkEvent(
                                type=NetworkEventType.ADDRESS_ADDED,
                                interface=interface,
                                key=key,
                                address=address,
                                address_family=family,
                            )
                            yield event

                await backend.wait_until_next_scan()

    async def removed_addresses(self) -> AsyncIterator[NetworkEvent]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            event objects describing the network addresses that were removed and
            the corresponding network interfaces
        """
        async for event in self.events():
            if event.type == NetworkEventType.ADDRESS_REMOVED:
                yield event

    async def removed_interfaces(self) -> AsyncIterator[NetworkEvent]:
        """Runs the network event detection in an asynchronous task.

        Yields:
            network interface objects, one for each interface whose removal was
            detected. Added interfaces are not reported.
        """
        async for event in self.events():
            if event.type == NetworkEventType.INTERFACE_REMOVED:
                yield event.interface

    def resume(self) -> None:
        """Resumes the network event detector task after a suspension."""
        self._suspended -= 1
        if not self._suspended and self._resume_event:
            self._resume_event.set()

    def suspend(self) -> None:
        """Temporarily suspends the network event detector."""
        self._suspended += 1
        if self._suspended and not self._resume_event:
            self._resume_event = Event()

    @contextmanager
    def suspended(self) -> Iterator[None]:
        """Async context manager that suspends the network event detector while the
        execution is in the context.
        """
        self.suspend()
        try:
            yield
        finally:
            self.resume()

    @staticmethod
    def _compare_addresses(
        entry: NetworkEventDetectorEntry, new: Dict[int, List[Any]]
    ) -> Tuple:
        """Compares a set of old addresses with a set of new addresses for a
        network interface, and reports the differences.

        Parameters:
            entry: the entry in the state object of the network event detector
                that contains the old addresses we have found at the last scan
            new: a dictionary mapping address families to the new addresses

        Returns:
            a tuple consisting of the list of added addresses and the list of
            removed addresses. Each entry in these lists consists of four
            items: interface, key of interface, address family and address
        """
        old = entry.addresses

        added, removed = [], []

        for family in set(list(old.keys()) + list(new.keys())):
            old_addresses = old.get(family)
            new_addresses = new.get(family)

            if old_addresses == new_addresses:
                # Nothing changed, good
                pass
            elif old_addresses is None:
                # The whole address family was added
                assert new_addresses is not None
                added.extend(
                    (entry.interface, entry.key, family, address)
                    for address in new_addresses
                )
            elif new_addresses is None:
                # The whole address family was removed
                removed.extend(
                    (entry.interface, entry.key, family, address)
                    for address in old_addresses
                )
            else:
                old_addresses, new_addresses = set(old_addresses), set(new_addresses)
                added.extend(
                    (entry.interface, entry.key, family, address)
                    for address in new_addresses - old_addresses
                )
                removed.extend(
                    (entry.interface, entry.key, family, address)
                    for address in old_addresses - new_addresses
                )

        return added, removed

    @staticmethod
    def _preprocess_params(params: Dict) -> Dict:
        """Preprocesses the parameters passed to the constructor.

        Currently a no-op.

        Parameters:
            params: the dictionary to process. It will be modified in-place.

        Returns:
            the same dictionary that was passed in
        """
        return params



from anyio import Condition, create_task_group, from_thread, to_thread
from contextlib import asynccontextmanager
from functools import partial
from typing import AsyncIterator, Optional

from .portable import PortableNetworkEventDetectorBackend

__all__ = ("SystemConfigurationBasedNetworkEventDetectorBackend",)


class SystemConfigurationBasedNetworkEventDetectorBackend(
    PortableNetworkEventDetectorBackend
):
    """Specialized network event detector backend for macOS using the
    SystemConfiguration framework.

    This backend works similarly to the PortableNetworkEventDetectorBackend_,
    but it uses an SCDynamicStore instance to detect changes to network
    interfaces and addresses instead of polling the network configuration.

    This backend is preferred by the autodetection mechanism on macOS over the
    default portable backend.
    """

    _condition: Optional[Condition]

    def __init__(self) -> None:
        super().__init__()
        self._condition = None

    @asynccontextmanager
    async def use(self) -> AsyncIterator[None]:
        self._ensure_condition_exists()
        async with create_task_group() as tg:
            tg.start_soon(
                partial(to_thread.run_sync, self._run_worker_thread, cancellable=True)
            )
            yield

    def _ensure_condition_exists(self) -> None:
        """Creates the condition object that the backend will use to notify
        listeners about network change events.

        This function is used to avoid having to create a condition object
        in the constructor, which may be executed outside the context of an
        event loop.
        """
        if self._condition is None:
            self._condition = Condition()

    def _on_network_changed(self, *args, **kwds):
        """Callback that is called by the SystemConfiguration framework when
        the network configuration has changed.

        This function runs in the context of the worker thread.
        """
        from_thread.run(self._on_network_changed_main)

    async def _on_network_changed_main(self):
        """Task that is scheduled on the main event loop when the network
        configuration has changed.

        This function runs in the context of the main event loop.
        """
        assert self._condition is not None
        async with self._condition:
            self._condition.notify_all()

    def _run_worker_thread(self):
        """Runs the worker thread that waits for network events."""
        from Foundation import (
            CFRunLoopAddSource,
            CFRunLoopGetCurrent,
            kCFRunLoopCommonModes,
            CFRunLoopRun,
        )
        from SystemConfiguration import (
            SCDynamicStoreCreate,
            SCDynamicStoreSetNotificationKeys,
            SCDynamicStoreCreateRunLoopSource,
        )

        store = SCDynamicStoreCreate(
            None, "global-network-watcher", self._on_network_changed, None
        )
        SCDynamicStoreSetNotificationKeys(
            store, None, ["State:/Network/Global/IPv4", "State:/Network/Global/IPv6"]
        )

        CFRunLoopAddSource(
            CFRunLoopGetCurrent(),
            SCDynamicStoreCreateRunLoopSource(None, store, 0),
            kCFRunLoopCommonModes,
        )

        CFRunLoopRun()

    async def wait_until_next_scan(self) -> None:
        assert self._condition is not None

        self._ensure_condition_exists()
        async with self._condition:
            await self._condition.wait()



from .task import NetworkEventDetector

from trio import run


async def main():
    detector = NetworkEventDetector()
    async for event in detector.events():
        print("{0!r}".format(event))


if __name__ == "__main__":
    run(main)



"""Portable network event detector backend based on `netifaces` and
continuous polling.
"""

from anyio import sleep
from typing import Any, Dict, List

from .base import NetworkEventDetectorBackend, NetworkInterface


class PortableNetworkEventDetectorBackend(NetworkEventDetectorBackend):
    """Portable network event detector backend based on `netifaces` and
    continuous polling.
    """

    def __init__(self):
        """Constructor."""
        import netifaces

        self._netifaces = netifaces

    async def get_addresses(self, interface: NetworkInterface) -> Dict[int, List[Any]]:
        addresses = self._netifaces.ifaddresses(interface)
        return {
            family: [item.get("addr") for item in addr_for_family]
            for family, addr_for_family in addresses.items()
        }

    def key_of(self, interface: NetworkInterface) -> str:
        return str(interface)

    async def scan(self) -> List[NetworkInterface]:
        return self._netifaces.interfaces()

    async def wait_until_next_scan(self) -> None:
        await sleep(1)



import platform

from pathlib import Path

from .base import NetworkEventDetectorBackend


def choose_backend() -> NetworkEventDetectorBackend:
    """Chooses an appropriate network event detector backend for the current
    platform when invoked with no arguments.

    Returns:
        a newly constructed network event detector backend instance
    """
    if platform.system() == "Linux" and Path("/proc/net/netlink").exists():
        from .netlink import NetlinkBasedNetworkEventDetectorBackend

        return NetlinkBasedNetworkEventDetectorBackend()
    elif platform.system() == "Darwin":
        from .macos import SystemConfigurationBasedNetworkEventDetectorBackend

        return SystemConfigurationBasedNetworkEventDetectorBackend()
    else:
        from .portable import PortableNetworkEventDetectorBackend

        return PortableNetworkEventDetectorBackend()



from anyio import sleep, wait_socket_readable
from os import getpid

from .portable import PortableNetworkEventDetectorBackend


RTMGRP_LINK = 1
RTMGRP_IPV4_IFADDR = 0x10
RTMGRP_IPV6_IFADDR = 0x100
RTMGRP_DECnet_IFADDR = 0x1000


class NetlinkBasedNetworkEventDetectorBackend(PortableNetworkEventDetectorBackend):
    """Specialized network event detector backend for Linux with netlink support.

    This backend works similarly to the PortableNetworkEventDetectorBackend_,
    but it uses a netlink socket to detect changes to network interfaces and
    addresses instead of polling the network configuration.

    This backend is preferred by the autodetection mechanism on Linux over the
    default portable backend if it detects that netlink support is enabled in
    the kernel.
    """

    def __init__(self):
        """Constructor."""
        super().__init__()

        from socket import socket, AF_NETLINK, SOCK_RAW, NETLINK_ROUTE

        self._socket = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)
        self._socket.bind(
            (
                getpid(),
                RTMGRP_LINK
                | RTMGRP_IPV4_IFADDR
                | RTMGRP_IPV6_IFADDR
                | RTMGRP_DECnet_IFADDR,
            )
        )
        self._socket.setblocking(False)

    def __del__(self):
        self._socket.close()

    async def wait_until_next_scan(self) -> None:
        await wait_socket_readable(self._socket)
        while True:
            try:
                data = self._socket.recv(4096)
                if not data:
                    break
            except BlockingIOError:
                break
            await sleep(0)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aio_net_events', 'aio_net_events.backends']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3.0.0,<4.0.0', 'netifaces>=0.10.9']

extras_require = \
{':sys_platform == "darwin"': ['pyobjc-core>=6.2.2,<7.0.0',
                               'pyobjc-framework-Cocoa>=6.2.2,<7.0.0',
                               'pyobjc-framework-SystemConfiguration>=6.2.2,<7.0.0']}

setup_kwargs = {
    'name': 'aio-net-events',
    'version': '1.0.3',
    'description': 'Asynchronous network configuration event detector for Python 3.7 and above',
    'long_description': '# aio-net-events\n\n`aio-net-events` is a Python library that provides asynchronous generators\nyielding events when the network configuration of the machine changes.\nCurrently only network interface additions / removals and IP address additions /\nremovals are supported; more events may be added later.\n\nSupports Windows, Linux and macOS at the moment.\n\nRequires Python >= 3.7.\n\nWorks with [`asyncio`](https://docs.python.org/3/library/asyncio.html) and\n[`trio`](https://trio.readthedocs.io/en/stable/).\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install\n`aio-net-events`.\n\n```bash\npip install aio-net-events\n```\n\n## Usage\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to\ndiscuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Tamas Nepusz',
    'author_email': 'tamas@collmot.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ntamas/aio-net-events/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .backends.autodetect import choose_backend
from .backends.base import NetworkInterface
from .task import NetworkEventDetector

__all__ = ("choose_backend", "NetworkEventDetector", "NetworkInterface")
